# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_patch_thread():
    import gevent.monkey

    gevent.monkey.patch_thread()


def test_patch_ssl_recursion():
    import gevent.monkey

    gevent.monkey.patch_all()

    # Delay imports and test creation of contexts from both bundled and installed urllib3
    import urllib3.util.ssl_

    import newrelic.packages.urllib3.util.ssl_

    context1 = urllib3.util.ssl_.create_urllib3_context()
    context2 = newrelic.packages.urllib3.util.ssl_.create_urllib3_context()
